from .settings_loader import load_settings


# Load config when `mizuhara` is imported
try:
    settings = load_settings()

# add exception handler to avoid exception during creating project or app
except FileNotFoundError:
    pass


"""
2025.10.26: change designate column type from Annotated to Optional and raw type.
"""

# import packages from default or pip library
from datetime import datetime, timezone
from typing_extensions import Optional
from uuid import UUID, uuid4
from sqlmodel import (SQLModel,
                      Field)

# import packages from this framework


# define Class for Common SQLModel
class IdMixin(SQLModel):
    id: int = Field(primary_key=True, default=None)

    def __setattr__(self, name, value) -> None:
        if name == "id":
            if hasattr(self, name) and getattr(self, name) is not None:
                raise ValueError(f"[Warning] it is not allowed to edit value of column '{name}' in '{self.__class__}' directly.")

        super().__setattr__(name, value)
        return None


class UUIDMixin(SQLModel):
    id: UUID = Field(primary_key=True, default=None)

    def __setattr__(self, name, value) -> None:
        if name == "id":
            if hasattr(self, name) and getattr(self, name) is not None:
                raise ValueError(f"[Warning] it is not allowed to edit value of column '{name}' in '{self.__class__}' directly.")

        super().__setattr__(name, value)
        return None


class RegisterDateTimeMixin(SQLModel):
    register_dt: datetime = Field(nullable=False, default=None)

    def __setattr__(self, name, value) -> None:
        if name == "register_dt":
            if hasattr(self, name) and getattr(self, name) is not None:
                raise ValueError(f"[Warning] it is not allowed to edit value of column '{name}' in '{self.__class__}' directly.")

        super().__setattr__(name, value)
        return None


class UpdateDateTimeMixin(SQLModel):
    update_dt: Optional[datetime] = Field(nullable=True, default=None)

    def renew_update_dt(self) -> None:
        super().__setattr__(name="update_dt", value=datetime.now(tz=timezone.utc))
        return None


# define function to control SQL event
# please follow the procedure below
# use decorator 'event.listens_for()' for event handling function
# event handler function must have 3 args: mapper, connection and target)
# in your model file, import listens_for from 'sqlalchemy.event'
# event.listens_for(target=TARGET_TABLE_CLASS_NAME, identifier=['before_insert', 'before_update',...])
def _assign_register_datetime(mapper, connection, target) -> None:
    if target.register_dt is None:
        target.register_dt = datetime.now(tz=timezone.utc)
    return None

def _assign_update_datetime(mapper, connection, target) -> None:
    target.renew_update_dt()
    return None

def _assign_uuid(mapper, connection, target) -> None:
    if target.id is None:
        target.id = uuid4()
    return None


# import packages from default or pip library
from datetime import datetime, timedelta, timezone
from mongoengine import Document, DateTimeField, StringField

# import packages from this framework
from settings import AUTHENTICATION


# settings
TOKEN_VALID_TIME: int = AUTHENTICATION.get("token").get("token_valid_time")

# define classes
class AccessTokenMixin(Document):
    meta = {'abstract': True}

    access_token: str = StringField(primary_key=True,
                                    null=False,
                                    unique=True)


class RefreshTokenMixin(Document):
    meta = {'abstract': True}

    refresh_token: str = StringField(null=False,
                                     unique=True)


class TokenValidDateTime(Document):
    """
    This class is for token management.
    it can be stored on your DB as a whitelist.
    """

    meta = {'abstract': True}

    issued_dt: datetime = DateTimeField(null=False)
    expiration_dt: datetime = DateTimeField(null=False)

    def __setattr__(self, name: str, value: datetime):
        if name in ["issued_at", "expiration_dt",]:
            if hasattr(self, name) and getattr(self, name) is not None:
                raise ValueError(f"[Warning] it is not allowed to edit value of column '{name}' in '{self.__class__}' directly.")

        super().__setattr__(name, value)
        return None

    def set_token_datetime(self) -> None:
        if self.issued_dt is None:
            super().__setattr__(name="issued_at", value=datetime.now(tz=timezone.utc))

        if self.expiration_dt is None:
            super().__setattr__(name="expiration_dt", value=datetime.now(tz=timezone.utc) + timedelta(minutes=TOKEN_VALID_TIME))

        return None


class JWTTokens(RefreshTokenMixin, AccessTokenMixin):
    meta = {'abstract': True}

    token_type: str = StringField(null=False,
                                  default="Bearer")


# define function to control MongoDB document event
# please follow the procedure below
# create event handler function with name starting with _.
# event handler function must have 3 args: sender, document, **kwargs)
# each db column can get from 'document'
# import pre_init or pre_save from 'mongoengine.signals' in your model file.
# pre_init.connect(EVNET_HANDLER_FUNC_NAME, sender=TABLE_CLASS_NAME)
def _fill_token_datetime_field(sender, document, **kwargs) -> None:
    document.set_token_datetime()
    return None



# import packages from default or pip library
from typing_extensions import Annotated, Doc
from pydantic import BaseModel


# import packages from this framework



# define Class for Common SQLModel
class DefaultJWTTokenView(BaseModel):
    """
    This class is a view template for returning JWT tokens information.
    Read Only.
    """

    access_token: Annotated[str,
    Doc("This is an access token string for authenticated user")]
    refresh_token: Annotated[str,
    Doc("This is a token string for reissuing access token")]
    token_type: Annotated[str,
    Doc("Token type that used in this class. Default is 'Bearer'")] = "Bearer"



# import packages from default or pip library
from pycountry import languages as pylanguage
from mongoengine import Document, BooleanField, IntField, StringField
from fastapi.exceptions import RequestValidationError

# import packages from this framework


# define Class for Common SQLModel
class AllowMaxAccessSettingMixin(Document):
    meta = {'abstract': True}

    allow_max_access: str = IntField(null=False, default=5, min_value=3, max_value=5)


class LanguageSettingMixin(Document):
    meta = {'abstract': True}

    language: str = StringField(null=False)

    # "assign 'alpha_2' or 'alpha_3'. default is 'alpha_2'"
    _alpha_code_type: str = "alpha_2"


class DarkModeSettingMixin(Document):
    meta = {'abstract': True}

    is_dark_mode: bool = BooleanField(null=False, default=False)


# define function to control MongoDB document event
# please follow the procedure below
# create event handler function with name starting with _.
# event handler function must have 3 args: sender, document, **kwargs)
# each db column can get from 'document'
# import pre_init or pre_save from 'mongoengine.signals' in your model file.
# pre_init.connect(EVNET_HANDLER_FUNC_NAME, sender=TABLE_CLASS_NAME)
def _convert_language_to_alpha2(sender, document, **kwargs):
    if document.language and len(document.language) != 2:
        try:
            document.language = pylanguage.get(name=document.language).alpha_2
        except AttributeError:
            raise RequestValidationError(errors={"input": "nation", "msg": "please input official country name."})

    return None

def _convert_language_to_alpha3(sender, document, **kwargs):
    if document.language and len(document.language) != 3:
        try:
            document.language = pylanguage.get(name=document.language).alpha_3
        except AttributeError:
            raise RequestValidationError(errors={"input": "nation", "msg": "please input official country name."})

    return None



import importlib.util
import os
import sys

CONFIG_NAME = "settings.py"

def find_config():
    """Search for settings.py in the user's project."""
    for path in sys.path:
        config_path = os.path.join(path, CONFIG_NAME)

        if os.path.exists(config_path):
            return config_path
    return None

def load_settings():
    """Dynamically import config.py if it exists."""
    settings_path = find_config()
    if not settings_path:
        raise FileNotFoundError("Not found 'settings.py' file. Please start projects first with command 'jknife'.")

    # create a module specification for config.py with naming config
    # importlib.util.spec_from_file_location(module name, import path)
    # type = ModuleSpec
    spec = importlib.util.spec_from_file_location("settings", settings_path)

    # create an empty module class instance from specification.
    # importlib.util.module_from_spec(spec)
    settings = importlib.util.module_from_spec(spec)

    # execute an empty module
    # load and execute config.py from config
    spec.loader.exec_module(settings)

    # Inject settings into `sys.modules` as `jknife.settings`
    sys.modules["jknife.settings"] = settings
    return settings


"""
2025.10.26: change designate column type from Annotated to Optional and raw type.
"""

# import packages from default or pip library
import hashlib
from datetime import datetime, timezone
from sqlmodel import SQLModel, Field

# import packages from this framework
from settings import PASSWORD_POLICIES


# password encryption algorithm
ENCRYPT_TYPE: str = PASSWORD_POLICIES.get("encrypt_type")


# define Class for Common SQLModel
class LastSigninDateTimeMixin(SQLModel):
    last_signin_dt: datetime = Field(nullable=True, default=None)

    def __setattr__(self, name, value) -> None:
        if name == "last_signin_dt":
            if hasattr(self, name) and getattr(self, name) is not None:
                raise ValueError(f"[Warning] it is not allowed to edit value of column '{name}' in '{self.__class__}' directly.")

        super().__setattr__(name, value)
        return None

    def update_signin_dt(self) -> None:
        super().__setattr__(name="last_signin_dt", value=datetime.now(tz=timezone.utc))
        return None


class SigninFailMixin(SQLModel):
    is_active: bool = Field(nullable=False, default=False)
    signin_fail: int = Field(nullable=False, default=0, ge=0, le=5)

    def activate_user(self) -> None:
        super().__setattr__(name="is_active", value=True)
        self.__init_signin_fail_count()
        return None

    def deactivate_user(self) -> None:
        super().__setattr__(name="is_active", value=False)
        return None

    def add_signin_fail_count(self, limit: int) -> None:
        if self.signin_fail < limit:
            super().__setattr__(name="signin_fail", value=self.signin_fail + 1)

            if self.signin_fail == limit:
                self.deactivate_user()

        return None

    def __init_signin_fail_count(self) -> None:
        super().__setattr__(name="signin_fail", value=0)
        return None


class IsAdminMixin(SQLModel):
    is_admin: bool = Field(nullable=False, default=False)

    def grant_admin(self) -> None:
        super().__setattr__(name="is_admin", value=True)
        return None

    def revoke_admin(self) -> None:
        super().__setattr__(name="is_admin", value=False)
        return None


class PasswordMixin(SQLModel):
    password: str = Field(nullable=False, unique=False)

    def check_password(self, password: str) -> bool:
        return self.password == encrypt_password(password=password)

    def change_password(self, old_pwd: str, new_pwd: str) -> bool:
        if self.password != encrypt_password(password=old_pwd):
            return False

        super().__setattr__(name="password", value=encrypt_password(password=new_pwd))
        return True

    def encrypt_password(self, enc_type:str=ENCRYPT_TYPE) -> None:
        super().__setattr__(name="password", value=encrypt_password(password=self.password, enc_type=enc_type))
        return None


class UsernameMixin(SQLModel):
    username: str = Field(nullable=False, unique=True)


# define function for users
def encrypt_password(password:str, enc_type:str=ENCRYPT_TYPE) -> str:
    if enc_type not in hashlib.algorithms_available:
        raise ValueError(f"'{enc_type}' is not supported hash method in this application. Password will be encrypted with default hash.")

    return getattr(hashlib, enc_type)(string=password.encode("utf-8")).hexdigest()


# define function to control SQL event
# please follow the procedure below
# use decorator 'event.listens_for()' for event handling function
# event handler function must have 3 args: mapper, connection and target)
# in your model file, import listens_for from 'sqlalchemy.event'
# event.listens_for(target=TARGET_TABLE_CLASS_NAME, identifier=['before_insert', 'before_update',...])




# import packages from default or pip library
from typing_extensions import Annotated, Doc
from pydantic import BaseModel

# import packages from this framework


# Error Mixin Class
class ErrMsgMixin(BaseModel):
    msg: Annotated[str,
    Doc("message for error")]


class ErrMsgTypeMixin(BaseModel):
    type: Annotated[str,
    Doc("customised message type. for example, 'password_mismatch' or 'unauthorised access'.")]


class ErrMsgBoolResultMixin(BaseModel):
    result: Annotated[bool, Doc("result of calling API in bool type.")] = False


class FieldValidationErrMixin(BaseModel):
    field: Annotated[str,
    Doc("field name which has an error.")]


# Error Model
class DefaultErrorMsgView(ErrMsgMixin):
    pass


class ErrMsgWithTypeView(ErrMsgTypeMixin, DefaultErrorMsgView):
    pass


class ErrMsgWithTypeAndResultView(ErrMsgBoolResultMixin, ErrMsgWithTypeView):
    pass


"""
2025.10.26: change designate column type from Annotated to Optional and raw type.
"""

# import packages from default or pip library
from pycountry import languages as pylanguage
from sqlmodel import SQLModel, Field

# import packages from this framework


# define Class for Common SQLModel
class AllowMaxAccessSettingMixin(SQLModel):
    allow_max_access: str = Field(nullable=False, default=5, ge=3, le=5)


class LanguageSettingMixin(SQLModel):
    language: str = Field(nullable=False)


class DarkModeSettingMixin(SQLModel):
    is_dark_mode: str = Field(nullable=False, default=False)


# define function to control SQL event
# please follow the procedure below
# use decorator 'event.listens_for()' for event handling function
# event handler function must have 3 args: mapper, connection and target)
# in your model file, import listens_for from 'sqlalchemy.event'
# event.listens_for(target=TARGET_TABLE_CLASS_NAME, identifier=['before_insert', 'before_update',...])
def _convert_language_to_alpha2(mapper, connection, target) -> None:
    if target.language is not None:
        target.language = pylanguage.get(name=target.language).alpha_2
    return None

def _convert_language_to_alpha3(mapper, connection, target) -> None:
    if target.language is not None:
        target.language = pylanguage.get(name=target.language).alpha_3
    return None


"""
2025.10.26: change designate column type from Annotated to Optional and raw type.
"""

# import packages from default or pip library
from sqlmodel import SQLModel, Field

# import packages from this framework


# define Class for Common SQLModel
class AllowIPsMixin(SQLModel):
    allow_ips: list[str] = Field(nullable=False,
                                 default=["127.0.0.1"])


# define function to control SQL event
# please follow the procedure below
# use decorator 'event.listens_for()' for event handling function
# event handler function must have 3 args: mapper, connection and target)
# in your model file, import listens_for from 'sqlalchemy.event'
# event.listens_for(target=TARGET_TABLE_CLASS_NAME, identifier=['before_insert', 'before_update',...])
def _convert_ips_to_str(mapper, connection, target) -> None:
    if target.allow_ips is not None:
        target.allow_ips = [ str(ip) for ip in target.allow_ips ]
    return None


# import packages from default or pip library
import jwt
from uuid import UUID
from redis.exceptions import ConnectionError
from random import randint
from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPBasic, HTTPBearer, HTTPAuthorizationCredentials
from datetime import datetime, timedelta, timezone

# import packages from this framework
from settings import AUTHENTICATION
from jknife.db import logger
from jknife.db import redis_connector


# create secret key in this file.
def create_secretkey() -> str:
    return "".join([chr(randint(65, 122)) for _ in range(256)])


# SETTINGS
_DEFAULT_TOKEN_VALID_TIME: int = 20
_DEFAULT_TOKEN_REFRESH_TIME: int = 24 * 60 * 30
_TOKEN_SETTING: dict = AUTHENTICATION.get("token")
_TOKEN_SECRET_KEY: str = _TOKEN_SETTING.get("secret_key") or create_secretkey()
_TOKEN_ISSUER: str = _TOKEN_SETTING.get("token_issuer") or "127.0.0.1"
_TOKEN_AUDIENCE: str = _TOKEN_SETTING.get("token_audience") or "http://127.0.0.1"
_TOKEN_EXP_ACCESS: int = _TOKEN_SETTING.get("token_valid_time") or _DEFAULT_TOKEN_VALID_TIME
_TOKEN_EXP_REFRESH: int = _TOKEN_SETTING.get("token_refresh_time") or _DEFAULT_TOKEN_REFRESH_TIME
_TOKEN_ALGORITHM: str = _TOKEN_SETTING.get("algorithm") or "HS256"


# Bearer Token Window
oauth2_scheme = HTTPBearer(scheme_name="bearer",
                           description="Authentication related with JWT token.\nInput access token got after signing in.",
                           auto_error=False)
basic_scheme = HTTPBasic(scheme_name="basic",
                         description="Authentication with username and password.\nInput username and password.",
                         auto_error=False)


# define function for tokens
def _check_blacklist_access_token(token: str) -> str:
    """
    check whether the access token is registered as blacklist in redis.

    :param token: access_token in string
    :return: access_token, if it is not registered in redis
    """

    user_unique = decode_jwt_token(token=token).get("sub")

    if redis_connector.get(name=f"access_token:{token}"):
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                            detail={"msg": "invalid access token"})

    elif not redis_connector.get(name=f"refresh_token:{user_unique}"):
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                            detail={"msg": "you have to sign in first."})

    return token

def check_whitelist_refresh_token(user_unique: str, token: str):
    """
    check whether the refresh token is registered as whitelist in redis.

    :param token: refresh_token in string
    :return: refresh_token, if it is registered in redis
    """

    registered_token: bytes = redis_connector.get(name=f"refresh_token:{user_unique}")
    if registered_token is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                            detail={"msg": "you have to sign in first."})

    if registered_token.decode("utf-8") != token:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                            detail={"msg": "invalid refresh token"})
    return token

def create_jwt_token(user_unique: str | UUID | int, user_data: dict, is_access: bool=True) -> str:
    """
    create jwt token(access and refresh) with user data.

    :param user_unique: primary key to distinguish specific object among the users.
    :param user_data: user data that is customised by developer.
    :param is_access: set token type as bool for access(default) or refresh(False).
    :return: encrypted string token.
    """

    payload: dict = {
        "iss": _TOKEN_ISSUER,
        "sub": str(user_unique),
        "aud": _TOKEN_AUDIENCE,
        "exp": datetime.now(timezone.utc) + timedelta(minutes=_TOKEN_EXP_ACCESS if is_access else _TOKEN_EXP_REFRESH),
        "user_data": user_data,
    }
    return jwt.encode(payload=payload,
                      key=_TOKEN_SECRET_KEY,
                      algorithm=_TOKEN_ALGORITHM)

def decode_jwt_token(token:str, is_access:bool=True) -> dict | None:
    """
    decode jwt token information that encrypted in token string.

    :param token: token value in string
    :param is_access: set token type as bool for access(default) or refresh(False).
    :return:
    """

    try:
        return jwt.decode(jwt=token,
                          key=_TOKEN_SECRET_KEY,
                          audience=_TOKEN_AUDIENCE,
                          algorithms=_TOKEN_ALGORITHM)

    except jwt.exceptions.ExpiredSignatureError:
        token_type: str = "access" if is_access else "refresh"
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                            detail={"msg": f"{token_type} token was expired"})

    except jwt.exceptions.DecodeError:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                            detail={"msg": "invalid access token. can not decode token information."})

def get_bearer_access_token(user_cred: HTTPAuthorizationCredentials = Depends(oauth2_scheme)) -> str:
    """
    get access token from HTTPBearer credential in RestAPI Docs Page.

    :param user_cred: get access_token information from HTTPBearer credential
    :return: return access_token
    """

    if user_cred is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                            detail={"msg": "missing access token in headers."})

    if user_cred.scheme.lower() != "bearer":
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                            detail={"msg": "invalid access token. token type is not match."})

    return _check_blacklist_access_token(token=user_cred.credentials)

def get_token_info(token: str, is_access: bool = True) -> dict:
    """
    decrypt JWT token and return its data as dict type.

    :param token: JWT token(access or refresh)
    :param is_access: set token type as bool for access(default) or refresh(False).
    :return: token information in dict type.
    """

    payload: dict = decode_jwt_token(token=token, is_access=is_access)
    return payload

def register_blacklist_access_token(token: str, ex: int=_TOKEN_EXP_ACCESS) -> None:
    """
    register JWT access token in redis after signout or changing password
    template: access_token: {ACCESS_TOKEN}

    :param token: access_token that will be registered in redis.
    :param ex: token expiration time in second.
    :return: None
    """

    try:
        redis_connector.set(name=f"access_token:{token}", value="1", ex=ex * 60)
    except ConnectionError:
        logger.warning(msg="fail to register access_token as a blacklist. can not connect to redis db.")

    return None

def register_whitelist_refresh_token(user_unique: str | UUID | int, token: str) -> None:
    """
    register JWT refresh token in redis after signin.
    template: refresh_token:{USER_UNIQUE}

    :param user_unique: primary key for each user.
    :param token: refresh token that was issued after signin.
    :return: None
    """
    try:
        redis_connector.set(name=f"refresh_token:{str(user_unique)}", value=token, ex=_TOKEN_EXP_REFRESH * 60)
    except ConnectionError:
        logger.warning(msg="fail to register refresh_token as a whitelist. can not connect to redis db.")

    return None

def remove_whitelist_refresh_token(user_unique: str | UUID | int) -> None:
    """
    remove registered refresh token in redis after signout or changing password.

    :param user_unique: primary key for each user.
    :return: None
    """

    try:
        redis_connector.delete(f"refresh_token:{str(user_unique)}")
    except ConnectionError:
        logger.warning(msg="fail to remove refresh token from redis. can not connect to redis db.")

    return None


# import packages from default or pip library
from mongoengine import Document, ListField

# import packages from this framework


# define Class for Common SQLModel
class AllowIPsMixin(Document):
    meta = {'abstract': True}

    allow_ips: list[str] = ListField(null=False,
                                     default=["127.0.0.1"])


# define function to control MongoDB document event
# please follow the procedure below
# create event handler function with name starting with _.
# event handler function must have 3 args: sender, document, **kwargs)
# each db column can get from 'document'
# import pre_init or pre_save from 'mongoengine.signals' in your model file.
# pre_init.connect(EVNET_HANDLER_FUNC_NAME, sender=TABLE_CLASS_NAME)
def _convert_ips_to_str(sender, document, **kwargs) -> None:
    if document.allow_ip:
        document.allow_ip = [ str(ip) for ip in document.allow_ip ]
    return None



# import packages from default or pip library
from datetime import datetime, timezone
from mongoengine import Document, BooleanField, DateTimeField, IntField, StringField

# import packages from this framework
from settings import PASSWORD_POLICIES
from ..rdbms.users import encrypt_password


# password encryption algorithm
ENCRYPT_TYPE: str = PASSWORD_POLICIES.get("encrypt_type")

# define Class for Common SQLModel
class LastSigninDateTimeMixin(Document):
    meta = {'abstract': True}

    last_signin_dt: datetime = DateTimeField(null=True, default=None)

    def __setattr__(self, name: str, value: datetime):
        if name == "last_signin_dt":
            if hasattr(self, name) and getattr(self, name) is not None:
                raise ValueError(f"[Warning] it is not allowed to edit value of column '{name}' in '{self.__class__}' directly.")

        super().__setattr__(name, value)
        return None

    def update_signin_dt(self):
        super().__setattr__(name="last_signin_dt", value=datetime.now(tz=timezone.utc))
        return None


class SigninFailMixin(Document):
    meta = {'abstract': True}

    is_active: bool = BooleanField(null=False, default=False, db_field="is_active")
    signin_fail: int = IntField(null=False, default=0, ge=0, le=5, db_field="signin_fail")

    def activate_user(self) -> None:
        super().__setattr__(name="is_active", value=True)
        self.__init_signin_fail_count()
        return None

    def deactivate_user(self) -> None:
        super().__setattr__(name="is_active", value=False)
        return None

    def add_signin_fail_count(self, limit: int) -> None:
        if self.signin_fail < limit:
            super().__setattr__(name="signin_fail", value=self.signin_fail + 1)

            if self.signin_fail == limit:
                self.deactivate_user()

        return None

    def __init_signin_fail_count(self) -> None:
        super().__setattr__(name="signin_fail", value=0)
        return None


class IsAdminMixin(Document):
    meta = {'abstract': True}

    is_admin: bool = BooleanField(null=False, default=False, db_field="is_admin")

    def grant_admin(self) -> None:
        super().__setattr__(name="is_admin", value=True)
        return None

    def revoke_admin(self) -> None:
        super().__setattr__(name="is_admin", value=False)
        return None


class PasswordMixin(Document):
    meta = {'abstract': True}

    password: str = StringField(null=False, unique=False)

    def __setattr__(self, name, value) -> None:
        if name == "password":
            if hasattr(self, name) and getattr(self, name) is not None:
                raise ValueError(f"[Warning] it is not allowed to edit value of column '{name}' in '{self.__class__}' directly.")

        super().__setattr__(name, value)
        return None

    def check_password(self, password: str) -> bool:
        return self.password == encrypt_password(password=password)

    def change_password(self, old_pwd: str, new_pwd: str) -> bool:
        if self.password != encrypt_password(password=old_pwd):
            return False

        super().__setattr__(name="password", value=encrypt_password(password=new_pwd))
        return True

    def encrypt_password(self, enc_type:str=ENCRYPT_TYPE) -> None:
        super().__setattr__(name="password", value=encrypt_password(password=self.password, enc_type=enc_type))
        return None


class UsernameMixin(Document):
    meta = {'abstract': True}

    username: str = StringField(null=False, unique=True)

    def __setattr__(self, name, value) -> None:
        if name == "username":
            if hasattr(self, name) and getattr(self, name) is not None:
                raise ValueError(f"[Warning] it is not allowed to edit value of column '{name}' in '{self.__class__}' directly.")

        super().__setattr__(name, value)
        return None


# define function to control MongoDB document event
# please follow the procedure below
# create event handler function with name starting with _.
# event handler function must have 3 args: sender, document, **kwargs)
# each db column can get from 'document'
# import pre_init or pre_save from 'mongoengine.signals' in your model file.
# pre_init.connect(EVNET_HANDLER_FUNC_NAME, sender=TABLE_CLASS_NAME)


# import packages from default or pip library
from datetime import datetime
from pydantic import BaseModel
from typing_extensions import Annotated, Doc, Optional
from uuid import UUID

# import packages from this framework


# define mixin class
class IdViewMixin(BaseModel):
    id: Annotated[int,
    Doc("Default integer id for each table row.")]


class UUIDViewMixin(BaseModel):
    id: Annotated[UUID,
    Doc("UUID format id for each table row.")]


class RegisterDateTimeViewMixin(BaseModel):
    register_dt: Annotated[Optional[datetime],
    Doc("Datetime that the row was added at.")]


class UpdateDateTimeViewMixin(BaseModel):
    update_dt: Annotated[Optional[datetime],
    Doc("Datetime that the row was updated at.")]


# import packages from default or pip library
from fastapi.exceptions import RequestValidationError

# import packages from this framework
from settings import PASSWORD_POLICIES


# settings
MIN_LENGTH: int = PASSWORD_POLICIES.get("min_length")
SPECIAL_CHARS: str = "!@#$%^&*()_-+=~`\'\";:\\|<>,.?/"


# define password policy functions
def password_minlen(password: str, field: str = "password"):
    if len(password) >= MIN_LENGTH:
        return password

    raise RequestValidationError(errors={"input": field,
                                         "msg": f"{field} must be at least 8 chars"})


def password_upchar(password: str, field: str = "password") -> str:
    for c in password:
        if c.isupper():
            return password

    raise RequestValidationError(errors={"input": field,
                                         "msg": f"{field} must contain at least one upper character"})

def password_lowchar(password: str, field: str = "password") -> str:
    for c in password:
        if c.islower():
            return password

    raise RequestValidationError(errors={"input": field,
                                         "msg": f"{field} must contain at least one lower character"})


def password_number(password: str, field: str = "password") -> str:
    for c in password:
        if c.isnumeric():
            return password

    raise RequestValidationError(errors={"input": field,
                                         "msg": f"{field} must contain at least one number"})


def password_special(password: str, field: str = "password") -> str:
    for c in password:
        if c in SPECIAL_CHARS:
            return password

    raise RequestValidationError(errors={"input":field,
                                         "msg":f"{field} must contain at least one special character"})


# define dependencies functions
def validate_password_policy(password: str, field: str = "password") -> str:
    policy_mapping: dict = {
        "PASSWORD_MINLEN": password_minlen,
        "PASSWORD_UPCHAR": password_upchar,
        "PASSWORD_LOWCHAR": password_lowchar,
        "PASSWORD_NUMBER": password_number,
        "PASSWORD_SPECIAL": password_special,
    }
    validate_list: list = PASSWORD_POLICIES.get("compliance")

    if len(validate_list) == 0:
        return password

    for func in validate_list:
        tmp = policy_mapping.get(func)
        if tmp is not None and callable(tmp):
            tmp(password=password, field=field)

    return password



import logging
import logging.config
from os import mkdir, listdir
from typing_extensions import Annotated, Doc
from settings import LOG_SETTINGS, DEBUG_MODE


# create logs folder to store log files.
if "logs" not in listdir():
    mkdir("logs")

# set logging config from settings.py
# apply default key:value for dictConfig
LOG_SETTINGS.update({"version": 1, "disable_existing_loggers": False})
logging.config.dictConfig(config=LOG_SETTINGS)


# define LoggerMgmt for multiple logging
class LoggerMgmt:
    """
    This class is designed to implement multiple logging system in FastAPI.

    :param logger_names: assign the logger name that you want to use in specific area.
                         you can define it in settings.py with LOG_SETTINGS and *_LOGGER_LIST
    """

    def __init__(self,
                 logger_names: Annotated[list[str],
                                         Doc("assign the name of loggers to record activities")]):
        self.__loggers = [logging.getLogger(name=name) for name in logger_names]
        if DEBUG_MODE:
            uvicorn_logger = logging.getLogger(name="uvicorn")
            uvicorn_logger.setLevel(level="DEBUG")
            self.__loggers.append(uvicorn_logger)

    def critical(self, msg: Annotated[str,
                                      Doc("log message for critical level")]) -> None:
        for logger in self.__loggers:
            logger.critical(msg=msg)

        return None

    def error(self, msg: Annotated[str,
                                   Doc("log message for error level")]) -> None:
        for logger in self.__loggers:
            logger.error(msg=msg)

        return None

    def warning(self, msg: Annotated[str,
                                     Doc("log message for warning level")]) -> None:
        for logger in self.__loggers:
            logger.warning(msg=msg)

        return None

    def info(self, msg: Annotated[str,
                                  Doc("log message for info level")]) -> None:
        for logger in self.__loggers:
            logger.info(msg=msg)

        return None

    def debug(self, msg: Annotated[str,
                                  Doc("log message for debug level")]) -> None:
        for logger in self.__loggers:
            logger.debug(msg=msg)

        return None


# import packages from default or pip library
from datetime import datetime, timezone
from typing_extensions import deprecated, Optional
from uuid import UUID, uuid4
from mongoengine import Document, UUIDField, DateTimeField


# import packages from this framework


# define Class for Common MongoDB
@deprecated("MongoDB does not requires to set uuid manually.")
class UUIDMixin(Document):
    meta = {'abstract': True}

    id: UUID = UUIDField(primary_key=True, default=None, unique=True)


class RegisterDateTimeMixin(Document):
    meta = {'abstract': True}

    register_dt: datetime = DateTimeField(null=False, default=None, db_field="register_dt")

    def __setattr__(self, name: str, value: datetime):
        if name == "register_dt":
            if hasattr(self, name) and getattr(self, name) is not None:
                raise ValueError(f"[Warning] it is not allowed to edit value of column '{name}' in '{self.__class__}' directly.")

        super().__setattr__(name, value)
        return None


class UpdateDateTimeMixin(Document):
    meta = {'abstract': True}

    update_dt: Optional[datetime] = DateTimeField(null=False, default=None, db_field="update_dt")

    def renew_update_dt(self) -> None:
        super().__setattr__(name="update_dt", value=datetime.now(tz=timezone.utc))
        return None


# define function to control MongoDB document event
# please follow the procedure below
# create event handler function with name starting with _.
# event handler function must have 3 args: sender, document, **kwargs)
# each db column can get from 'document'
# import pre_init or pre_save from 'mongoengine.signals' in your model file.
# pre_init.connect(EVNET_HANDLER_FUNC_NAME, sender=TABLE_CLASS_NAME)
@deprecated("MongoDB does not requires to set uuid manually.")
def _assign_uuid(sender, document, **kwargs) -> None:
    if document.id is None:
        document.id = uuid4()
    return None

def _assign_register_datetime(sender, document, **kwargs) -> None:
    if document.register_dt is None:
        document.register_dt = datetime.now(tz=timezone.utc)
    return None

def _assign_update_datetime(sender, document, **kwargs) -> None:
    document.renew_update_dt()
    return None



"""
2025.10.26: change designate column type from Annotated to Optional and raw type.
"""

# import packages from default or pip library
from datetime import date
from sqlmodel import SQLModel, Field
from fastapi.exceptions import RequestValidationError
from pycountry import countries as pycountries

# import packages from this framework


# define Class for Common SQLModel
class AddressMixin(SQLModel):
    address: str = Field(nullable=False)


class BirthdateMixin(SQLModel):
    birthdate: date = Field(nullable=False)


class CPNumMixin(SQLModel):
    cp_num: str = Field(nullable=False)


class EmailMixin(SQLModel):
    email: str = Field(nullable=False, unique=True)


class FirstnameMixin(SQLModel):
    firstname: str = Field(nullable=False, index=True)


class LastnameMixin(SQLModel):
    lastname: str = Field(nullable=False, index=True)


class NationMixin(SQLModel):
    nation: str = Field(nullable=False)


class NicknameMixin(SQLModel):
    nickname: str = Field(unique=True, index=True)


class PostalCodeMixin(SQLModel):
    postal_code: str = Field(nullable=False)


# define function to control SQL event
# please follow the procedure below
# use decorator 'event.listens_for()' for event handling function
# event handler function must have 3 args: mapper, connection and target)
# in your model file, import listens_for from 'sqlalchemy.event'
# event.listens_for(target=TARGET_TABLE_CLASS_NAME, identifier=['before_insert', 'before_update',...])
def _convert_country_to_alpha2(mapper, connection, target):
    if target.nation is not None and len(target.nation) != 2:
        try:
            target.nation = pycountries.get(name=target.nation).alpha_2
        except AttributeError:
            raise RequestValidationError(errors={"input": "nation", "msg": "please input official country name."})
    return None

def _convert_country_to_alpha3(mapper, connection, target):
    if target.nation is not None and len(target.nation) != 3:
        try:
            target.nation = pycountries.get(name=target.nation).alpha_3
        except AttributeError:
            raise RequestValidationError(errors={"input": "nation", "msg": "please input official country name."})
    return None

def _lower_firstname(mapper, connection, target) -> None:
    if target.firstname is not None:
        target.firstname = target.firstname.lower()
    return None

def _lower_lastname(mapper, connection, target) -> None:
    if target.lastname is not None:
        target.lastname = target.lastname.lower()
    return None




# import packages from default or pip library
from datetime import date
from mongoengine import Document, DateField, StringField, EmailField
from fastapi.exceptions import RequestValidationError
from pycountry import countries as pycountries

# import packages from this framework


# define Class for Common SQLModel
class AddressMixin(Document):
    meta = {'abstract': True}

    address: str = StringField(null=False)


class BirthdateMixin(Document):
    meta = {'abstract': True}

    birthdate: date = DateField(null=False)


class CPNumMixin(Document):
    meta = {'abstract': True}

    cp_num: str = StringField(null=False)


class EmailMixin(Document):
    meta = {'abstract': True}

    email: str = EmailField(null=False, unique=True)


class FirstnameMixin(Document):
    meta = {'abstract': True}

    firstname: str = StringField(null=False)


class LastnameMixin(Document):
    meta = {'abstract': True}

    lastname: str = StringField(null=False)


class NationMixin(Document):
    meta = {'abstract': True}

    nation: str = StringField(null=False)


class NicknameMixin(Document):
    meta = {'abstract': True}

    nickname: str = StringField(unique=True)


class PostalCodeMixin(Document):
    meta = {'abstract': True}

    postal_code: str = StringField(null=False)


# define function to control MongoDB document event
# please follow the procedure below
# create event handler function with name starting with _.
# event handler function must have 3 args: sender, document, **kwargs)
# each db column can get from 'document'
# import pre_init or pre_save from 'mongoengine.signals' in your model file.
# pre_init.connect(EVNET_HANDLER_FUNC_NAME, sender=TABLE_CLASS_NAME)
def _convert_country_to_alpha2(sender, document, **kwargs):
    if document.nation and len(document.nation) != 2:
        try:
            document.nation = pycountries.get(name=document.nation).alpha_2
        except AttributeError:
            raise RequestValidationError(errors={"input": "nation", "msg": "please input official country name."})

    return None

def _convert_country_to_alpha3(sender, document, **kwargs):
    if document.nation and len(document.nation) != 3:
        try:
            document.nation = pycountries.get(name=document.nation).alpha_3
        except AttributeError:
            raise RequestValidationError(errors={"input": "nation", "msg": "please input official country name."})

    return None

def _lower_firstname(sender, document, **kwargs) -> None:
    if document.firstname:
        document.firstname = document.firstname.lower()
    return None

def _lower_lastname(sender, document, **kwargs) -> None:
    if document.lastname:
        document.lastname = document.lastname.lower()
    return None



# import packages from default or pip library
from datetime import date
from typing_extensions import Annotated, Doc
from pydantic import BaseModel, EmailStr

# import packages from this framework


# define mixin class
class BirthdateViewMixin(BaseModel):
    birthdate: Annotated[date,
    Doc("user's birthdate")]


class CPNumViewMixin(BaseModel):
    cp_num: Annotated[str,
    Doc("cellphone number.")]


class EmailInputViewMixin(BaseModel):
    email: Annotated[EmailStr,
    Doc("email address for user. it can be replaced username.")]


class FirstNameViewMixin(BaseModel):
    firstname: Annotated[str,
    Doc("firstname of user")]


class LastNameViewMixin(BaseModel):
    lastname: Annotated[str,
    Doc("lastname of user")]


class NationViewMixin(BaseModel):
    nation: Annotated[str,
    Doc("the nation that user came from.")]


class PostalCodeViewMixin(BaseModel):
    postal_code: Annotated[str,
    Doc("postal number of address.")]



import sys
import os
import subprocess


FOLDER_LIST: tuple = ("models", "routers", "views", )

# Help MSG for commands 'jknife'
HELP_MSG = """
[ Command 'jknife' ]
Usage: jknife [sub_command] [3rd argument]

* sub_command:
- startproject [PROJECT_NAME]   : create project filesystem
- createapp [APP_NAME]          : create API application in project
- runserver [OPTIONS]           : run server with uvicorn

* options
[runserver OPTIONS]
--host [HOSTNAME]               : run server with hostname [HOSTNAME] 
--port [PORT_INT]               : run server with portnumber [PORT]
--reload                        : reload uvicorn server automatically after editing source.
"""

# Template for APP_NAME.py in each folder.
MODULE_IMPORT_STRING_DICT: dict = {
    "models": """# import packages from default or pip library
from typing_extensions import Annotated, Doc, Optional

# import packages from this framework below


# set CONSTANT variables below.


# define your own customising class below
# please be advised that you can not use Annotated and Doc with RDBMS to define data type in table class.
# Do not forget to assign 'table=True' arguments in table class when you use RDBMS.

""",
    "routers": """# import packages from default or pip library
from fastapi import APIRouter, Depends, status, HTTPException
from typing_extensions import Annotated, Doc

# import packages from this framework below


# set CONSTANT variables below.
router = APIRouter(prefix="/{}",
                   tags=['{}'])

# define your own customising API function below
# Please note that you have to import db_connector from main.py to get RDBMS SQL session when you use RDBMS. 

""",
    "views": """# import packages from default or pip library
from pydantic import BaseModel, field_validator, ValidationError

# import packages from this framework below


# set CONSTANT variables below.


# define your own customizing class below
"""
}

# WARN MSGs
WARN_MSG_ALREADY_EXIST_PROJECT: str = """
[WARNING] You already have project named '{}'.
"""
WARN_MSG_ALREADY_EXIST_APP: str = """
[WARNING] You already have app named '{}'.
"""

# ERROR MSGs
ERROR_MSG_START_PROJECT: str = """
[ERROR] 
- You have to execute 'createapp' command in your project folder.

* Command for Starting Project: jknife startproject [PROJECT_NAME]
"""
ERROR_MSG_IMPOSSIBLE_RUNSERVER_OUT_OF_PROJECT: str = """
[ERROR] You have to execute 'runserver' command in your project folder.
"""

# File Contents
CONTENTS_IN_SETTINGS: str = """# this file is charge of config for custom fastapi.
from os import environ


# [ DEBUG ]
# this variable is a mode switcher.
# if you are about to publish your application to public, please turn this switch to False.
# on the other hands, during development, please switch this value to True.
# DEBUG_MODE provides
# - query logs on database.
DEBUG_MODE: bool = True

# [ API_VERSION ]
# API_VERSION is a management variables for your api.
# This value will be used to call api as an URL prefix. e.g) /api/{YOUR_API_VERSION}/
# you can modify the value of API_VERSION to whatever you want.
API_VERSION: str = "v1"

# [ DATABASE ]
# DATABASE is charge of connection to database server.
# recommend to store DB information in env var and get them by using environ.get()
# this framework supports some databases.
# - supported type: ["none", "postgresql", "mysql", "sqlite", "mongo"]
#
# type 'none': if you do not want to use database. it is a default if the type is None
# structure of database
# - type in ['postgresql', 'mysql', 'mongo']
#   - key: "dbms"
#   - values: {"username": DB_USERNAME, "password": DB_PASSWORD, "host": DB_HOST, "port": DB_PORT, "db": DB_NAME}
#   -         if you use mongo, it is possible to add key 'authSource' for name of database that is responsible for auth.
# - type 'sqlite'
#   - sqlite_filepath: input path of sqlite3 database file in an absolute way or a relative one.
# "redis" exist for the Redis - cached db.
DATABASE: dict = {
    "type": environ.get("db_type"),
    "dbms": {
        "username": environ.get("db_username"),
        "password": environ.get("db_password"),
        "host": environ.get("db_host"),
        "port": int(environ.get("db_port")),
        "db": environ.get("db_name"),
        "authSource": environ.get("mongo_auth_source")
    },
    "sqlite": {
        "sqlite_filepath": "./fastapi.db"
    },
    "redis": {
        "host": environ.get("redis_host"),
        "port": environ.get("redis_port"),
    }
}

# [ AUTHENTICATION ]
# AUTHENTICATION is charge of JWT token for authenticated user.
# token: information for JWT token
# - token_issuer: token issuer
# - token_audience: IP or URL address that allows this token
# - token_valid_time: interval time between when the access token issued and expired in minutes.
# - token_refresh_time: interval time in minute for reissuing access token.
# - secret_key: secret key to encrypt token
# - algorithm: select encryption algorithm
#   * HS: for HMAC symmetric algorithm
#   * RS: for RSA symmetric algorithm
#   * ES: for ECDSA asymmetric algorithm
AUTHENTICATION: dict = {
    "token": {
        "token_issuer": "http://127.0.0.1",
        "token_audience": "http://127.0.0.1",
        "token_valid_time": 20,
        "token_refresh_time": 60 * 24 * 30,
        "secret_key": "THIS IS A TEST",
        "algorithm": "HS256"
    }
}

# [ PASSWORD_POLICIES ]
# PASSWORD_POLICIES is charge of password compliance, when users are created or change their password
# compliance: the list of password policies
# - PASSWORD_MINLEN: limit the minimum length of password. default is 8 but it can be customized with 'min_length'.
# - PASSWORD_UPCHAR: password must contain at least one upper alphabet character.
# - PASSWORD_LOWCHAR: password must contain at least one lower alphabet character.
# - PASSWORD_NUMBER: password must contain at least one numeric character.
# - PASSWORD_SPECIAL: password must contain at least one special character.
# encrypt_type: algorithm for encrypting password. please refer to 'hashlib.algorithms_available'
# min_length: minimum length of password. it will be used for PASSWORD_MINLEN
PASSWORD_POLICIES: dict = {
    "compliance": [
        "PASSWORD_MINLEN",
        "PASSWORD_UPCHAR",
        "PASSWORD_LOWCHAR",
        "PASSWORD_NUMBER",
        "PASSWORD_SPECIAL"
    ],
    "encrypt_type": "sha256",
    "min_length": 8
}


# [ LOG_SETTINGS ]
# LOG_SETTINGS is based on python dictConfig
# (https://docs.python.org/3/library/logging.config.html#dictionary-schema-details).
# Refer to the official documentation, write down your own config for logging.
# {"version": 1, "disable_existing_logger": False} will be applied automatically.
# If DEBUG_MODE is True, uvicorn log will be printed out on your console.
# Default Logger:
# * DB: The logger 'db' is charge of database logging.
#       if you want to change the name of logger, you should also change the DB_LOGGER_LIST below.
#       or you can not see any proper log.
LOG_SETTINGS: dict = {
    "formatters": {
        "default": {
            "format": "[%(asctime)s] %(name)s:%(levelname)s - %(message)s",
            "datefmt": "%Y-%m-%d %H:%M:%S:%s"
        }
    },
    "handlers": {
        "main": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "INFO",
            "formatter": "default",
            "filename": "logs/all_logs.log",
            "maxBytes": 1024 * 1024 * 100,          # 10 MB
            "backupCount": 3
        },
        "db": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "INFO",
            "formatter": "default",
            "filename": "logs/db.log",
            "maxBytes": 1024 * 1024 * 100,          # 10 MB
            "backupCount": 3
        },
        "db_error": {
            "class": "logging.handlers.RotatingFileHandler",
            "level": "ERROR",
            "formatter": "default",
            "filename": "logs/db_error.log",
            "maxBytes": 1024 * 1024 * 100,           # 10 MB
            "backupCount": 3
        }
    },
    "loggers": {
        "": {
            "level": "INFO",
            "handlers": ["main"],
        },
        "db": {
            "level": "INFO",
            "handlers": ["db", "db_error"],
            "propagate": True,
        },
    }
}

# [ LOGGER_LIST ]
# create your own logger instance with core.logging.LoggerMgmt and use it in specific realm.
DB_LOGGER_LIST = ("db",)


### END CONTENTS ###
"""
CONTENTS_IN_MAIN: str = """# import modules from python library
import importlib, os
from fastapi import FastAPI, Request, status
from fastapi.responses import JSONResponse
from fastapi.exceptions import RequestValidationError


# import required objects from projects
from jknife.db import DBConnector
from settings import API_VERSION


# set main url for all router
MAIN_URL: str = f"/api/{API_VERSION}"


# DB Connector
db_connector = DBConnector()

# start application
app = FastAPI(lifespan=db_connector.startup_db)


# add exception handler for ValidationException
@app.exception_handler(exc_class_or_status_code=RequestValidationError)
async def validation_error(req: Request, exc: RequestValidationError):
    errors: dict = exc.errors()[0] if isinstance(exc.errors(), list) else dict(exc.errors())
    result: dict = {"field": errors.get("input"), "msg": errors.get("msg")}
    return JSONResponse(
        status_code=status.HTTP_400_BAD_REQUEST,
        content={"detail": result}
    )


# add routers_bak
for router in os.listdir(os.path.abspath(path="routers")):
    if not router.startswith("__"):
        tmp_module = importlib.import_module(name=f"routers.{router.split('.')[0]}")
        app.include_router(router=tmp_module.router,
                           prefix=MAIN_URL)

"""


def startproject(pjt_name: str) -> None:
    # check whether the folder 'pjt_name' exist or not.
    if pjt_name in os.listdir():
        # print WARNING
        print(WARN_MSG_ALREADY_EXIST_PROJECT.format(pjt_name))
        return None

    # create project folder
    os.mkdir(pjt_name)

    # create working folders for jknife
    for folder_name in FOLDER_LIST:
        if f"{folder_name}" not in os.listdir(pjt_name):
            os.mkdir(f"{pjt_name}/{folder_name}")

            if "__init__.py" not in os.listdir(f"{pjt_name}/{folder_name}"):
                with open(f"{pjt_name}/{folder_name}/__init__.py", mode="w") as f:
                    f.write("")

    # create settings.py file
    if f"settings.py" not in os.listdir(pjt_name):
        with open(f"{pjt_name}/settings.py", "w") as f:
            f.write(CONTENTS_IN_SETTINGS)

    # create main.py file
    if f"main.py" not in os.listdir(pjt_name):
        with open(f"{pjt_name}/main.py", "w") as f:
            f.write(CONTENTS_IN_MAIN)

    return None

def createapp(app_name: str) -> None:
    # check main folders for jknife
    for folder_name in FOLDER_LIST:
        try:
            if not os.listdir(folder_name):
                with open(f"{folder_name}/__init__.py", "w") as f:
                    f.write("")

            if f"{app_name}.py" not in os.listdir(folder_name):
                with open(f"{folder_name}/{app_name}.py", "w") as f:
                    text = MODULE_IMPORT_STRING_DICT.get(folder_name)

                    if folder_name == "routers":
                        text = text.format(app_name, app_name)

                    f.write(text)

            else:
                raise AttributeError

        except FileNotFoundError:
            print(ERROR_MSG_START_PROJECT)
            return None

        except AttributeError:
            print(WARN_MSG_ALREADY_EXIST_APP.format(app_name))
            return None

    return None

def help_msg():
    print(HELP_MSG)
    return None

def run_server(options: list, app_name: str = "main"):
    for f in FOLDER_LIST:
        if f not in os.listdir():
            print(ERROR_MSG_IMPOSSIBLE_RUNSERVER_OUT_OF_PROJECT)
            return None

    args = " ".join([ option.replace("=", " ") for option in options ])
    command = f"uvicorn {app_name}:app {args}"
    result_cmd = subprocess.call(command, shell=True)
    return None

def main() -> None:
    read_command: list = sys.argv

    try:
        sub_command = read_command[1]

        if sub_command == "startproject":
            startproject(pjt_name=read_command[2])
            return None

        elif sub_command == "createapp":
            createapp(app_name=read_command[2])
            return None

        elif sub_command == "runserver":
            run_server(read_command[2:])
            return None

    except IndexError:
        pass

    help_msg()
    return None


if __name__ == "__main__":
    main()


# import packages from default or pip library
import importlib, os, redis
from contextlib import asynccontextmanager
from typing_extensions import Annotated, Doc
from fastapi import FastAPI
from sqlalchemy.engine.base import Engine
from sqlalchemy.exc import OperationalError
from sqlmodel import SQLModel, Session, create_engine
from mongoengine import connect

# import packages from this framework
from ..logging import LoggerMgmt
from settings import *


# Define supported database
SUPPORT_DATABASE: dict = {
    "none": {"type": "none", "engine": "none"},
    "postgresql": {"type": "dbms", "engine": "rdbms"},
    "mysql": {"type": "dbms", "engine": "rdbms"},
    "mongo": {"type": "dbms", "engine": "mongo"},
    "sqlite": {"type": "sqlite", "engine": "rdbms"},
}

# set logger
logger = LoggerMgmt(logger_names=DB_LOGGER_LIST or ["db"])


# define redis instance
redis_host: str = DATABASE.get("redis").get("host")
redis_port: int = DATABASE.get("redis").get("port")
redis_connector = redis.Redis(host=redis_host, port=redis_port)
if redis_host is not None:
    try:
        redis_connector.ping()
        if not DEBUG_MODE:
            redis_connector.flushdb()

    except redis.exceptions.ConnectionError:
        logger.warning(msg="Can not connect to Redis server. Please check your server's information.")


# define DBConnector
class DBConnector:
    """
    This class will make the application connect to database server which is designed in settings.py
    The list of supported database is written in SUPPORT_DATABASE above.
    """

    def __init__(self):
        self.__type: Annotated[str,
        Doc("Assign the database type. It must be one of value in SUPPORT_DB_TYPES.")] = DATABASE.get("type") or "none"
        self.__db_info: Annotated[dict | None,
        Doc(f"type of database defined in SUPPORTED_DATABASE")] = SUPPORT_DATABASE.get(self.__type)

        # import table classes in models for sqlmodel.
        if self.__db_info.get("engine") == "rdbms":
            for file in os.listdir(os.path.abspath("models")):
                if not file.startswith("__") and file.endswith(".py"):
                    importlib.import_module(f"models.{file.split('.')[0]}")

        # get connection information from settings.py
        if self.__db_info is not None:
            self.__conn_info: Annotated[dict | None,
            Doc("the type of database from DATABASE in settings.py")] = DATABASE.get(self.__db_info.get("type"))

            if self.__type in SUPPORT_DATABASE.keys() and self.__conn_info is not None:

                # for normal DBMS server
                if SUPPORT_DATABASE.get(self.__type).get("type") == "dbms":
                    self.__db: Annotated[str,
                    Doc("The name of database that will be used for this projects")] = self.__conn_info.get("db")
                    self.__host: Annotated[str,
                    Doc("The name of database that will be used for this projects")] = self.__conn_info.get("host")
                    self.__port: Annotated[str,
                    Doc("The name of database that will be used for this projects")] = self.__conn_info.get("port")
                    self.__username: Annotated[str,
                    Doc("The name of database that will be used for this projects")] = self.__conn_info.get("username")
                    self.__password: Annotated[str,
                    Doc("The name of database that will be used for this projects")] = self.__conn_info.get("password")
                    self.__auth_source: Annotated[str,
                    Doc("This class member is for mongo. Input the name of database that the username came from.")] = self.__conn_info.get("auth_source")

                # for sqlite3
                elif SUPPORT_DATABASE.get(self.__type).get("type") == "filedb":
                    self.__sqlite3_filepath: Annotated[str,
                    Doc("This class member is for sqlite3 only. Input must be absolute or relative path.")] = self.__conn_info.get("sqlite_filepath")

                self.__engine: Annotated[Engine,
                Doc("engine will be used to get db session.")] = self.__create_engine()


    def __create_engine(self) -> Engine | None:
        args: Annotated[dict,
        Doc("Arguments for create_engine()")] = {"check_same_thread": False} \
            if SUPPORT_DATABASE.get("type") == "filedb" else {}
        logger.debug(f"DB connection arguments: {args}")

        url = f"{self.__type}://{self.__username}:{self.__password}@{self.__host}:{self.__port}/{self.__db}" \
            if SUPPORT_DATABASE.get(self.__type).get("type") == "dbms" else f"{self.__type}:///{self.__sqlite3_filepath}"
        logger.debug(f"DB connection URL: {url}")

        if SUPPORT_DATABASE.get(self.__type).get("engine") == "rdbms":
            logger.info(f"creating RDBMS engine for '{self.__type}' database...")
            return create_engine(url=url,
                                 echo=DEBUG_MODE,
                                 connect_args=args)

        # create mongo client module.
        connect(**self.__conn_info)
        return None

    @asynccontextmanager
    async def startup_db(self, app: FastAPI):
        if self.__type == "none":
            logger.debug("There is no database config.")
            yield

        else:
            if self.__type in SUPPORT_DATABASE:
                try:
                    if SUPPORT_DATABASE.get(self.__type).get("engine") == "rdbms":
                        SQLModel.metadata.create_all(bind=self.__engine)
                        logger.info(f"successfully connected to '{self.__type}' database!")

                    else:
                        logger.info(f"successfully created `{self.__type}` connecting module!")

                    yield

                except (OperationalError,) as e:
                    logger.error(f"Can not connect to DB server: '{self.__host}:{self.__port}/{self.__db}'")
                    return

            else:
                logger.warning(f"'{self.__type} is not supported database type.")
                return

    # return session for RDBMS
    def get_session(self):
        with Session(self.__engine) as session:
            yield session



# import packages from default or pip library
from datetime import datetime
from typing_extensions import Annotated, Doc
from pydantic import BaseModel, field_validator

# import packages from this framework
from jknife.views.personnel_info import EmailInputViewMixin
from jknife.dependencies.users import validate_password_policy


# define mixin class
class UsernameInputViewMixin(BaseModel):
    username: Annotated[str,
    Doc("username")]


class InputPasswordViewMixin(BaseModel):
    password: Annotated[str,
    Doc("password for user")]


class LastSigninDateTimeViewMixin(BaseModel):
    last_signin_dt: Annotated[datetime | None,
    Doc("User's last signin datetime.")] = None


class SigninFailMixin(BaseModel):
    is_active: Annotated[bool,
    Doc("show whether the user is activated or not.")]
    signin_fail: Annotated[int,
    Doc("If the user fail to login, this value will be incremented.")]


# define common view class
class ChangePasswordView(InputPasswordViewMixin):
    new_password: Annotated[str,
    Doc("New password for changing password")]

    @field_validator('new_password', mode='before')
    @classmethod
    def check_new_password(cls, value) -> str:
        return validate_password_policy(password=value, field="new_password")


class DefaultSignupView(InputPasswordViewMixin, UsernameInputViewMixin):
    password: Annotated[str,
    Doc("password for user")]

    @field_validator('password', mode='before')
    @classmethod
    def check_password(cls, value) -> str:
        return validate_password_policy(password=value)


class DefaultEmailSignupView(InputPasswordViewMixin, EmailInputViewMixin):
    password: Annotated[str,
    Doc("password for user")]

    @field_validator('password', mode='before')
    @classmethod
    def check_password(cls, value) -> str:
        return validate_password_policy(password=value)


class UsernameSigninView(InputPasswordViewMixin, UsernameInputViewMixin):
    pass


class EmailSigninView(InputPasswordViewMixin, EmailInputViewMixin):
    pass



"""
2025.10.26: change designate column type from Annotated to Optional and raw type.
"""

# import packages from default or pip library
from datetime import datetime, timedelta, timezone
from sqlmodel import SQLModel, Field

# import packages from this framework
from settings import AUTHENTICATION


# settings
TOKEN_VALID_TIME: int = AUTHENTICATION.get("token").get("token_valid_time")

# define classes
class AccessTokenMixin(SQLModel):
    access_token: str = Field(primary_key=True,
                              nullable=False,
                              unique=True)


class RefreshTokenMixin(SQLModel):
    refresh_token: str = Field(nullable=False,
                               unique=True)


class TokenValidDateTimeMixin(SQLModel):
    """
    This class is for token management.
    it can be stored on your DB as a whitelist.
    """

    issued_dt: datetime= Field(nullable=False)
    expiration_dt: datetime = Field(nullable=False)

    def __setattr__(self, name: str, value: datetime):
        if name in ["issued_at", "expiration_dt",]:
            if hasattr(self, name) and getattr(self, name) is not None:
                raise ValueError(f"[Warning] it is not allowed to edit value of column '{name}' in '{self.__class__}' directly.")

        super().__setattr__(name, value)
        return None

    def set_token_datetime(self) -> None:
        if self.issued_dt is None:
            super().__setattr__(name="issued_at", value=datetime.now(tz=timezone.utc))

        if self.expiration_dt is None:
            super().__setattr__(name="expiration_dt", value=datetime.now(tz=timezone.utc) + timedelta(minutes=TOKEN_VALID_TIME))

        return None


class JWTTokens(RefreshTokenMixin, AccessTokenMixin):
    token_type: str = Field(nullable=False,
                            default="Bearer")


# define function to control SQL event
# please follow the procedure below
# use decorator 'event.listens_for()' for event handling function
# event handler function must have 3 args: mapper, connection and target)
# in your model file, import listens_for from 'sqlalchemy.event'
# event.listens_for(target=TARGET_TABLE_CLASS_NAME, identifier=['before_insert', 'before_update',...])
def _fill_token_datetime_field(mapper, connection, target) -> None:
    target.set_token_datetime()
    return None
